@@ -3,15 +3,16 @@
  *
  * This library is free software; you can redistribute it and/or modify
  * it under the terms of the GNU Lesser General Public License
- * version 3.0 as published by the Free Software Foundation.
+ * as published by the Free Software Foundation; either
+ * version 2.1 of the License, or (at your option) any later version.
  *
  * This library is distributed in the hope that it will be useful,
  * but WITHOUT ANY WARRANTY; without even the implied warranty of
  * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
- * GNU Lesser General Public License version 3.0 for more details.
+ * GNU Lesser General Public License for more details.
  *
  * You should have received a copy of the GNU Lesser General Public
- * License along with this library. If not, see
+ * License along with this library; if not, see
  * <http://www.gnu.org/licenses/>.
  *
  * Authored by: Michal Hruby <michal.mhr@gmail.com>
@@ -22,9 +23,7 @@
 #endif
 
 #include "zeitgeist-data-source-registry.h"
-#include "zeitgeist-eggdbusconversions.h"
 #include "zeitgeist-marshal.h"
-#include "eggzeitgeistbindings.h"
 
 /**
  * SECTION:zeitgeist-data-source-registry
@@ -47,16 +46,18 @@ G_DEFINE_TYPE (ZeitgeistDataSourceRegistry,
 
 typedef struct
 {
-  /* The connection to the ZG daemon
-   * Note: The EggZeitgeistDataSourceRegistry is owned by 
-   * the EggDBusObjectProxy!
-   */
-  EggDBusObjectProxy *registry_proxy;
-  EggZeitgeistDataSourceRegistry *registry;
+  /* Our connection to the bus */
+  GDBusConnection *connection;
 
-  gulong src_registered_id;
-  gulong src_disconnected_id;
-  gulong src_enabled_id;
+  /* The connection to the ZG daemon.
+   * If index != NULL it means we have a connection */
+  GDBusProxy *registry;
+
+  /* Method calls queued up while waiting for a proxy  */
+  GSList *method_dispatch_queue;
+
+  /* DBus signal handlers */
+  guint dbus_signals_id;
 
 } ZeitgeistDataSourceRegistryPrivate;
 
@@ -76,26 +77,46 @@ enum
 
   LAST_SIGNAL
 };
+
 static guint _registry_signals[LAST_SIGNAL] = { 0 };
 
+typedef struct
+{
+  ZeitgeistDataSourceRegistry *self;
+  const gchar                 *method_name;
+  GVariant                    *params;
+  GCancellable                *cancellable;
+  GAsyncReadyCallback          cb;
+  gpointer                     user_data;
+} MethodDispatchContext;
+
+static void    on_proxy_acquired (GObject *source_object,
+                                  GAsyncResult *res,
+                                  gpointer user_data);
+
+static void    dispatch_method         (MethodDispatchContext *ctx);
+
+static void    dispatch_async_callback (GObject               *source_object,
+                                        GAsyncResult          *res,
+                                        gpointer               user_data);
+
 static void
 zeitgeist_data_source_registry_init (ZeitgeistDataSourceRegistry *self)
 {
   ZeitgeistDataSourceRegistryPrivate *priv;
-  EggDBusConnection   *conn;
 
   priv = ZEITGEIST_DATA_SOURCE_REGISTRY_GET_PRIVATE (self);
-  
-  /* Set up the connection to the ZG daemon */  
-  conn = egg_dbus_connection_get_for_bus (EGG_DBUS_BUS_TYPE_SESSION);
-  priv->registry_proxy = 
-    egg_dbus_connection_get_object_proxy (conn,
-                                          "org.gnome.zeitgeist.Engine",
-                                          "/org/gnome/zeitgeist/data_source_registry");
-  
-  priv->registry =
-    EGG_ZEITGEIST_QUERY_INTERFACE_DATA_SOURCE_REGISTRY (priv->registry_proxy);
-  g_object_unref (conn);
+
+  /* Set up the connection to the ZG daemon */
+  g_dbus_proxy_new_for_bus (G_BUS_TYPE_SESSION,
+                            G_DBUS_PROXY_FLAGS_NONE,
+                            NULL,
+                            "org.gnome.zeitgeist.Engine",
+                            "/org/gnome/zeitgeist/data_source_registry",
+                            "org.gnome.zeitgeist.DataSourceRegistry",
+                            NULL,
+                            on_proxy_acquired,
+                            g_object_ref (self));
 }
 
 static void
@@ -107,27 +128,27 @@ zeitgeist_data_source_registry_finalize (GObject *object)
   registry = ZEITGEIST_DATA_SOURCE_REGISTRY (object);
   priv = ZEITGEIST_DATA_SOURCE_REGISTRY_GET_PRIVATE (registry);
 
-  if (priv->src_registered_id)
+  if (priv->dbus_signals_id)
     {
-      g_signal_handler_disconnect (priv->registry, priv->src_registered_id);
+      g_dbus_connection_signal_unsubscribe (priv->connection,
+                                            priv->dbus_signals_id);
+      priv->dbus_signals_id = 0;
     }
 
-  if (priv->src_disconnected_id)
+  if (priv->registry)
     {
-      g_signal_handler_disconnect (priv->registry, priv->src_disconnected_id);
-    }
-
-  if (priv->src_enabled_id)
-    {
-      g_signal_handler_disconnect (priv->registry, priv->src_enabled_id);
-    }
-
-  /* Note: priv->registry is owned by priv->registry_proxy */
-  if (priv->registry_proxy)
-    {
-      g_object_unref (priv->registry_proxy);
+      g_object_unref (priv->registry);
+      priv->registry = NULL;
     }
   
+  if (priv->connection)
+    {
+      g_object_unref (priv->connection);
+      priv->connection = NULL;
+    }
+
+
+
   G_OBJECT_CLASS (zeitgeist_data_source_registry_parent_class)->finalize (object);
 }
 
@@ -170,64 +191,47 @@ zeitgeist_data_source_registry_set_property (GObject      *object,
 }
 
 static void
-zeitgeist_data_source_registry_src_registered (
-    ZeitgeistDataSourceRegistry *registry,
-    EggZeitgeistDataSource *source)
-{
-  ZeitgeistDataSource *data_source;
-
-  data_source = _egg_zeitgeist_data_source_to_zeitgeist_data_source (source);
-
-  g_signal_emit (registry, _registry_signals[SOURCE_REGISTERED],
-                 0, data_source);
-
-  g_object_unref (data_source);
-}
-
-static void
-zeitgeist_data_source_registry_src_disconnected (
-    ZeitgeistDataSourceRegistry *registry,
-    EggZeitgeistDataSource *source)
-{
-  ZeitgeistDataSource *data_source;
-
-  data_source = _egg_zeitgeist_data_source_to_zeitgeist_data_source (source);
-
-  g_signal_emit (registry, _registry_signals[SOURCE_DISCONNECTED],
-                 0, data_source);
-
-  g_object_unref (data_source);
-}
-
-static void
-zeitgeist_data_source_registry_src_enabled (
-    ZeitgeistDataSourceRegistry *registry,
-    gchar *unique_id,
-    gboolean enabled)
-{
-  g_signal_emit (registry, _registry_signals[SOURCE_ENABLED],
-                 0, unique_id, enabled);
-}
-
-static void
-zeitgeist_data_source_registry_constructed (GObject *object)
+on_dbus_signal_received (GDBusConnection *connection,
+                         const gchar *sender_name,
+                         const gchar *object_path,
+                         const gchar *interface_name,
+                         const gchar *signal_name,
+                         GVariant *parameters,
+                         gpointer user_data)
 {
+  ZeitgeistDataSourceRegistry        *self;
   ZeitgeistDataSourceRegistryPrivate *priv;
+  ZeitgeistDataSource                *src;
+  gchar                              *unique_id;
+  gboolean                            enabled;
 
-  priv = ZEITGEIST_DATA_SOURCE_REGISTRY_GET_PRIVATE (object);
+  self = ZEITGEIST_DATA_SOURCE_REGISTRY (user_data);
+  priv = ZEITGEIST_DATA_SOURCE_REGISTRY_GET_PRIVATE (self);
 
-  priv->src_registered_id = g_signal_connect_swapped (
-      priv->registry, "data-source-registered",
-      G_CALLBACK (zeitgeist_data_source_registry_src_registered),
-      object);
-  priv->src_disconnected_id = g_signal_connect_swapped (
-      priv->registry, "data-source-disconnected",
-      G_CALLBACK (zeitgeist_data_source_registry_src_disconnected),
-      object);
-  priv->src_enabled_id = g_signal_connect_swapped (
-      priv->registry, "data-source-enabled",
-      G_CALLBACK (zeitgeist_data_source_registry_src_enabled),
-      object);
+  if (g_strcmp0 (signal_name, "DataSourceRegistered") == 0)
+    {
+      src = zeitgeist_data_source_new_from_variant (g_variant_get_child_value (parameters, 0)); // FIXME: ref?
+      g_signal_emit (self, _registry_signals[SOURCE_REGISTERED],
+                     0, src);
+      g_object_unref (src);
+    }
+  else if (g_strcmp0 (signal_name, "DataSourceDisconnected") == 0)
+    {
+      src = zeitgeist_data_source_new_from_variant (g_variant_get_child_value (parameters, 0)); // FIXME: ref?
+      g_signal_emit (self, _registry_signals[SOURCE_DISCONNECTED],
+                     0, src);
+      g_object_unref (src);
+    }
+  else if (g_strcmp0 (signal_name, "DataSourceEnabled") == 0)
+    {
+      g_variant_get (parameters, "sb", &unique_id, &enabled); // FIXME: unref?
+      g_signal_emit (self, _registry_signals[SOURCE_ENABLED],
+          0, unique_id, enabled);
+      g_free (unique_id);
+    }
+  else
+    g_warning ("Unknown signal from Zeitgeist Data Source Registry: %s",
+               signal_name);
 }
 
 static void
@@ -239,7 +243,6 @@ zeitgeist_data_source_registry_class_init (ZeitgeistDataSourceRegistryClass *kla
   object_class->finalize     = zeitgeist_data_source_registry_finalize;
   object_class->get_property = zeitgeist_data_source_registry_get_property;
   object_class->set_property = zeitgeist_data_source_registry_set_property;
-  object_class->constructed  = zeitgeist_data_source_registry_constructed;
 
   _registry_signals[SOURCE_REGISTERED] =
     g_signal_new ("source-registered",
@@ -259,6 +262,7 @@ zeitgeist_data_source_registry_class_init (ZeitgeistDataSourceRegistryClass *kla
                   g_cclosure_marshal_VOID__OBJECT,
                   G_TYPE_NONE, 1, ZEITGEIST_TYPE_DATA_SOURCE);
 
+  // FIXME: This signal should really be "source-toggled" and give ZgDataSrc as only arg
   _registry_signals[SOURCE_ENABLED] =
     g_signal_new ("source-enabled",
                   G_OBJECT_CLASS_TYPE (object_class),
@@ -272,27 +276,6 @@ zeitgeist_data_source_registry_class_init (ZeitgeistDataSourceRegistryClass *kla
                             sizeof (ZeitgeistDataSourceRegistryPrivate));
 }
 
-/* Used to marshal the async callbacks from EggDBus into ones
- * coming from this DataSourceRegistry instance */
-static void
-dispatch_async_callback (GObject      *source_object,
-                         GAsyncResult *res,
-                         gpointer      user_data)
-{ 
-  gpointer                    *data = (gpointer*) user_data;
-  ZeitgeistDataSourceRegistry *self = ZEITGEIST_DATA_SOURCE_REGISTRY (data[0]);
-  GAsyncReadyCallback          callback = (GAsyncReadyCallback) data[1];
-  gpointer                     _user_data = data[2];
-
-  if (callback != NULL)
-    {
-      callback (G_OBJECT (self), res, _user_data);
-    }
-  
-  g_object_unref (self);
-  g_free (user_data);
-}
-
 /*
  * API BELOW HERE
  */
@@ -327,24 +310,22 @@ zeitgeist_data_source_registry_get_data_sources (
                                      gpointer                     user_data)
 {
   ZeitgeistDataSourceRegistryPrivate *priv;
-  gpointer                           *dispatch_data;
+  MethodDispatchContext              *ctx;
 
   g_return_if_fail (ZEITGEIST_IS_DATA_SOURCE_REGISTRY (self));
   g_return_if_fail (cancellable == NULL || G_IS_CANCELLABLE (cancellable));
 
   priv = ZEITGEIST_DATA_SOURCE_REGISTRY_GET_PRIVATE (self);
 
-  dispatch_data = g_new (gpointer, 3);
-  dispatch_data[0] = g_object_ref (self);
-  dispatch_data[1] = callback;
-  dispatch_data[2] = user_data;
+  ctx = g_new0 (MethodDispatchContext, 1);
+  ctx->self = g_object_ref (self);
+  ctx->method_name = "GetDataSources";
+  ctx->params = g_variant_new ("()");
+  ctx->cancellable = cancellable;
+  ctx->cb = callback;
+  ctx->user_data = user_data;
 
-  egg_zeitgeist_data_source_registry_get_data_sources (
-      priv->registry,
-      EGG_DBUS_CALL_FLAGS_NONE,
-      cancellable,
-      dispatch_async_callback,
-      dispatch_data);
+  dispatch_method (ctx);
 }
 
 /**
@@ -364,33 +345,32 @@ zeitgeist_data_source_registry_get_data_sources_finish (
                                      GError                      **error)
 {
   ZeitgeistDataSourceRegistryPrivate *priv;
-  EggDBusArraySeq                    *array_seq;
-  gboolean                            call_result;
-  GPtrArray                          *data_sources = NULL;
+  GVariant                           *val, *vsources;
+  GPtrArray                          *sources;
 
   g_return_val_if_fail (ZEITGEIST_IS_DATA_SOURCE_REGISTRY (self), NULL);
+  g_return_val_if_fail (G_IS_ASYNC_RESULT (res), FALSE);
+  g_return_val_if_fail (error == NULL || *error == NULL, FALSE);
 
   priv = ZEITGEIST_DATA_SOURCE_REGISTRY_GET_PRIVATE (self);
 
-  call_result = egg_zeitgeist_data_source_registry_get_data_sources_finish (
-      priv->registry,
-      &array_seq,
-      res,
-      error);
+  val = g_dbus_proxy_call_finish (priv->registry, res, error);
 
-  data_sources =
-    _egg_zeitgeist_data_sources_to_zeitgeist_data_sources (array_seq);
-  g_ptr_array_set_free_func (data_sources, g_object_unref);
+  if (val == NULL)
+    return NULL;
 
-  g_object_unref (array_seq);
+  /* Unpack return value from wrapper struct */
+  vsources = g_variant_get_child_value (val, 0);
+  g_variant_unref (val);
 
-  return data_sources;
+  return zeitgeist_data_sources_from_variant (vsources); // owns ref
 }
 
 /**
  * zeitgeist_data_source_registry_register_data_source:
  * @self: Instance of #ZeitgeistDataSourceRegistry.
- * @source: Data source to register.
+ * @source: Data source to register. If this is a
+ *          floating reference it will be consumed
  * @cancellable: a #GCancellable or #NULL.
  * @callback: a GAsyncReadyCallback to call when the request is finished.
  * @user_data: the data to pass to callback function.
@@ -409,8 +389,8 @@ zeitgeist_data_source_registry_register_data_source (
     gpointer                     user_data)
 {
   ZeitgeistDataSourceRegistryPrivate *priv;
-  gpointer                           *dispatch_data;
-  EggDBusArraySeq                    *templates;
+  MethodDispatchContext              *ctx;
+  GVariant                           *vsource;
 
   g_return_if_fail (ZEITGEIST_IS_DATA_SOURCE_REGISTRY (self));
   g_return_if_fail (ZEITGEIST_IS_DATA_SOURCE (source));
@@ -418,30 +398,17 @@ zeitgeist_data_source_registry_register_data_source (
 
   priv = ZEITGEIST_DATA_SOURCE_REGISTRY_GET_PRIVATE (self);
 
-  g_object_ref_sink (G_OBJECT (source));
+  vsource = zeitgeist_data_source_to_variant (source); // sinks ref
   
-  dispatch_data = g_new (gpointer, 3);
-  dispatch_data[0] = g_object_ref (self);
-  dispatch_data[1] = callback;
-  dispatch_data[2] = user_data;
+  ctx = g_new0 (MethodDispatchContext, 1);
+  ctx->self = g_object_ref (self);
+  ctx->method_name = "RegisterDataSource";
+  ctx->params = vsource;
+  ctx->cancellable = cancellable;
+  ctx->cb = callback;
+  ctx->user_data = user_data;
 
-  templates = _zeitgeist_events_to_egg_zeitgeist_events (
-      zeitgeist_data_source_get_event_templates (source)
-  );
-
-  egg_zeitgeist_data_source_registry_register_data_source (
-      priv->registry,
-      EGG_DBUS_CALL_FLAGS_NONE,
-      zeitgeist_data_source_get_unique_id (source),
-      zeitgeist_data_source_get_name (source),
-      zeitgeist_data_source_get_description (source),
-      templates,
-      cancellable,
-      dispatch_async_callback,
-      dispatch_data);
-
-  g_object_unref (templates);
-  g_object_unref (source);
+  dispatch_method (ctx);
 }
 
 /**
@@ -459,19 +426,20 @@ zeitgeist_data_source_registry_register_data_source_finish (
     GError                     **error)
 {
   ZeitgeistDataSourceRegistryPrivate *priv;
-  gboolean                            result = FALSE;
+  GVariant                           *val;
 
   g_return_val_if_fail (ZEITGEIST_IS_DATA_SOURCE_REGISTRY (self), FALSE);
+  g_return_val_if_fail (G_IS_ASYNC_RESULT (res), FALSE);
+  g_return_val_if_fail (error == NULL || *error == NULL, FALSE);
 
   priv = ZEITGEIST_DATA_SOURCE_REGISTRY_GET_PRIVATE (self);
+  val = g_dbus_proxy_call_finish (priv->registry, res, error);
 
-  egg_zeitgeist_data_source_registry_register_data_source_finish (
-      priv->registry,
-      &result,
-      res,
-      error);
+  if (val == NULL)
+    return FALSE;
 
-  return result;
+  g_variant_unref (val);
+  return TRUE;
 }
 
 void
@@ -484,7 +452,9 @@ zeitgeist_data_source_registry_set_data_source_enabled (
     gpointer                     user_data)
 {
   ZeitgeistDataSourceRegistryPrivate *priv;
-  gpointer                           *dispatch_data;
+  MethodDispatchContext              *ctx;
+  GVariant                           *params;
+  GVariantBuilder                     b;
 
   g_return_if_fail (ZEITGEIST_IS_DATA_SOURCE_REGISTRY (self));
   g_return_if_fail (unique_id != NULL);
@@ -492,36 +462,140 @@ zeitgeist_data_source_registry_set_data_source_enabled (
 
   priv = ZEITGEIST_DATA_SOURCE_REGISTRY_GET_PRIVATE (self);
 
-  dispatch_data = g_new (gpointer, 3);
-  dispatch_data[0] = g_object_ref (self);
-  dispatch_data[1] = callback;
-  dispatch_data[2] = user_data;
+  /* Build the method params */
+  g_variant_builder_init (&b, G_VARIANT_TYPE ("(sb)"));
+  g_variant_builder_add (&b, "s", unique_id);
+  g_variant_builder_add (&b, "b", enabled);
+  params = g_variant_builder_end (&b);
 
-  egg_zeitgeist_data_source_registry_set_data_source_enabled (
-      priv->registry,
-      EGG_DBUS_CALL_FLAGS_NONE,
-      unique_id,
-      enabled,
-      cancellable,
-      dispatch_async_callback,
-      dispatch_data);
+  ctx = g_new0 (MethodDispatchContext, 1);
+  ctx->self = g_object_ref (self);
+  ctx->method_name = "SetDataSourceEnabled";
+  ctx->params = params;
+  ctx->cancellable = cancellable;
+  ctx->cb = callback;
+  ctx->user_data = user_data;
+
+  dispatch_method (ctx);
 }
 
-void
+gboolean
 zeitgeist_data_source_registry_set_data_source_enabled_finish (
     ZeitgeistDataSourceRegistry *self,
     GAsyncResult                *res,
     GError                     **error)
 {
   ZeitgeistDataSourceRegistryPrivate *priv;
+  GVariant                           *val;
 
-  g_return_if_fail (ZEITGEIST_IS_DATA_SOURCE_REGISTRY (self));
+  g_return_val_if_fail (ZEITGEIST_IS_DATA_SOURCE_REGISTRY (self), FALSE);
+  g_return_val_if_fail (G_IS_ASYNC_RESULT (res), FALSE);
+  g_return_val_if_fail (error == NULL || *error == NULL, FALSE);
 
   priv = ZEITGEIST_DATA_SOURCE_REGISTRY_GET_PRIVATE (self);
+  val = g_dbus_proxy_call_finish (priv->registry, res, error);
 
-  egg_zeitgeist_data_source_registry_set_data_source_enabled_finish (
-      priv->registry,
-      res,
-      error);
+  if (val == NULL)
+    return FALSE;
+
+  g_variant_unref (val);
+  return TRUE;
 }
 
+static void
+on_proxy_acquired (GObject *source_object,
+                   GAsyncResult *res,
+                   gpointer user_data)
+{
+  ZeitgeistDataSourceRegistry        *self;
+  ZeitgeistDataSourceRegistryPrivate *priv;
+  GError                             *error;
+
+  self = ZEITGEIST_DATA_SOURCE_REGISTRY (user_data);
+  priv = ZEITGEIST_DATA_SOURCE_REGISTRY_GET_PRIVATE (self);
+
+  error = NULL;
+  priv->registry = g_dbus_proxy_new_finish (res, &error);
+
+  if (error != NULL)
+    {
+      g_critical ("Unable to connect to Zeitgeist daemon: %s",
+                  error->message);
+      g_error_free (error);
+    }
+  else
+    {
+      /* Dispatch all method calls queued while waiting for the proxy */
+      priv->method_dispatch_queue = g_slist_reverse (priv->method_dispatch_queue);
+      g_slist_foreach (priv->method_dispatch_queue, (GFunc) dispatch_method, NULL);
+      g_slist_free (priv->method_dispatch_queue);
+      priv->method_dispatch_queue = NULL;
+    }
+
+  /* Grab a ref on the GDBusConnection as well */
+  priv->connection = G_DBUS_CONNECTION (g_object_ref (g_dbus_proxy_get_connection (priv->registry)));
+
+  /* Connect to all DBus signals from the data source registry */
+  priv->dbus_signals_id = g_dbus_connection_signal_subscribe (
+      priv->connection,
+      "org.gnome.zeitgeist.Engine",
+      "org.gnome.zeitgeist.DataSourceRegistry",
+      NULL, /* Listen for any signal on this interface */
+      "/org/gnome/zeitgeist/data_source_registry",
+      NULL,
+      G_DBUS_SIGNAL_FLAGS_NONE,
+      on_dbus_signal_received,
+      self, NULL);
+
+  /* Release ref we held during async op */
+  g_object_unref (self);
+}
+
+/* Send off the DBus method call, or queue it if we don't
+ * have a proxy at this point */
+static void
+dispatch_method (MethodDispatchContext *ctx)
+{
+  ZeitgeistDataSourceRegistryPrivate *priv;
+
+  priv = ZEITGEIST_DATA_SOURCE_REGISTRY_GET_PRIVATE (ctx->self);
+
+  if (priv->registry)
+    {
+      g_dbus_proxy_call (priv->registry,
+                         ctx->method_name,
+                         ctx->params,
+                         G_DBUS_CALL_FLAGS_NONE,
+                         -1,
+                         ctx->cancellable,
+                         dispatch_async_callback,
+                         ctx);
+    }
+  else
+    priv->method_dispatch_queue = g_slist_prepend (priv->method_dispatch_queue,
+                                                   ctx);
+}
+
+/* Used to marshal the async callbacks from GDBus into ones
+ * coming from this ZeitgeistIndex instance */
+static void
+dispatch_async_callback (GObject      *source_object,
+                         GAsyncResult *res,
+                         gpointer      user_data)
+{
+  GVariant              *var;
+  MethodDispatchContext *ctx = (MethodDispatchContext*) user_data;
+
+  if (ctx->cb != NULL)
+      ctx->cb (G_OBJECT (ctx->self), res, ctx->user_data);
+  else
+    {
+      /* Caller ignores response - finish the call our selves */
+      var = g_dbus_proxy_call_finish (G_DBUS_PROXY (source_object), res, NULL);
+      if (var != NULL)
+        g_variant_unref (var);
+    }
+
+  g_object_unref (ctx->self);
+  g_free (ctx);
+}
