@@ -3,15 +3,16 @@
  *
  * This library is free software; you can redistribute it and/or modify
  * it under the terms of the GNU Lesser General Public License
- * version 3.0 as published by the Free Software Foundation.
+ * as published by the Free Software Foundation; either
+ * version 2.1 of the License, or (at your option) any later version.
  *
  * This library is distributed in the hope that it will be useful,
  * but WITHOUT ANY WARRANTY; without even the implied warranty of
  * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
- * GNU Lesser General Public License version 3.0 for more details.
+ * GNU Lesser General Public License for more details.
  *
  * You should have received a copy of the GNU Lesser General Public
- * License along with this library. If not, see
+ * License along with this library; if not, see
  * <http://www.gnu.org/licenses/>.
  *
  * Authored by
@@ -279,8 +280,8 @@ zeitgeist_event_set_manifestation (ZeitgeistEvent *event,
  * Get the event actor. The actor represents the party responsible for
  * triggering the event. When the actor is an application
  * (which it almost always is) the actor is encoded in the
- * <emphasis>app://</emphasis> URI scheme with the base name of the .desktop
- * file for the application appended. Eg. <emphasis>app://firefox.desktop</emphasis>
+ * <emphasis>application://</emphasis> URI scheme with the base name of the .desktop
+ * file for the application appended. Eg. <emphasis>application://firefox.desktop</emphasis>
  *
  * Returns: A URI designating the actor of the event
  */
@@ -297,13 +298,13 @@ zeitgeist_event_get_actor (ZeitgeistEvent *event)
  * zeitgeist_event_set_actor:
  * @event: The event to set the actor for
  * @actor: URI designating the actor triggering the event.
- *         Fx. <emphasis>app://firefox.desktop</emphasis>
+ *         Fx. <emphasis>application://firefox.desktop</emphasis>
  *
  * Get the event actor. The actor represents the party responsible for
  * triggering the event. When the actor is an application
  * (which it almost always is) the actor is encoded in the
- * <emphasis>app://</emphasis> URI scheme with the base name of the .desktop
- * file for the application appended. Eg. <emphasis>app://firefox.desktop</emphasis>
+ * <emphasis>application://</emphasis> URI scheme with the base name of the .desktop
+ * file for the application appended. Eg. <emphasis>application://firefox.desktop</emphasis>
  */
 void
 zeitgeist_event_set_actor (ZeitgeistEvent *event,
@@ -566,7 +567,7 @@ zeitgeist_event_new (void)
  * Returns: A newly create #ZeitgeistEvent instance. The returned event will
  *          have a floating reference which will be consumed if you pass the
  *          event to any of the methods provided by this library. If you
- *          do not do that then you must free the event youself with
+ *          do not do that then you must free the event yourself with
  *          g_object_unref()
  */
 ZeitgeistEvent* 
@@ -625,3 +626,311 @@ zeitgeist_event_new_full_valist (const gchar *interpretation,
     
   return self;
 }
+
+/**
+ * zeitgeist_event_new_from_variant:
+ * @event: A #GVariant with signature defined in
+ *         #ZEITGEIST_EVENT_VARIANT_SIGNATURE. If @event is a floating
+ *         reference the floating reference will be consumed.
+ *
+ * Parse the data in a #GVariant and build a #ZeitgeistEvent from it.
+ * The reverse operation of this is zeitgeist_event_to_variant().
+ *
+ * Returns: A newly allocated #ZeitgeistEvent filled with the metadata,
+ *          subjects, and payload described by @event. The returned event will
+ *          have a floating reference which will be consumed if you pass the
+ *          event to any of the methods provided by this library. If you
+ *          do not do that then you must free the event yourself with
+ *          g_object_unref()
+ */
+ZeitgeistEvent*
+zeitgeist_event_new_from_variant (GVariant *event)
+{
+  ZeitgeistEvent   *result;
+  ZeitgeistSubject *subject;
+  GVariantIter     *event_data, *subjects, *payload_data, *subject_data;
+  gchar            *str;
+  GByteArray       *payload;
+  gint              payload_size;
+  guchar            payload_byte;
+
+  g_return_val_if_fail (event != NULL, NULL);
+
+  g_variant_ref_sink (event);
+
+  result = zeitgeist_event_new ();
+  g_variant_get (event, ZEITGEIST_EVENT_VARIANT_SIGNATURE,
+                 &event_data, &subjects, &payload_data);
+
+  /* Parse event data */
+  if (g_variant_iter_n_children (event_data) < 5)
+    {
+      g_critical ("Event data truncated at length %d",
+                  g_variant_iter_n_children (event_data));
+      goto cleanup;
+    }
+
+  // FIXME: Use g_variant_get_string() to cause less reallocations
+  g_variant_iter_loop (event_data, "s", &str);
+  zeitgeist_event_set_id (result, g_ascii_strtoull (str, NULL, 0));
+
+  g_variant_iter_loop (event_data, "s", &str);
+  zeitgeist_event_set_timestamp (result,  g_ascii_strtoll (str, NULL, 0));
+
+  g_variant_iter_loop (event_data, "s", &str);
+  zeitgeist_event_set_interpretation (result,  str[0] == '\0' ? NULL : str);
+
+  g_variant_iter_loop (event_data, "s", &str);
+  zeitgeist_event_set_manifestation (result,  str[0] == '\0' ? NULL : str);
+
+  g_variant_iter_loop (event_data, "s", &str);
+  zeitgeist_event_set_actor (result,  str[0] == '\0' ? NULL : str);
+
+  /* Build the list of subjects */
+  while (g_variant_iter_loop (subjects, "as", &subject_data))
+  {
+    /* Parse subject data */
+    if (g_variant_iter_n_children (subject_data) < 7)
+      {
+        g_critical ("Subject data truncated at length %d",
+                    g_variant_iter_n_children (subject_data));
+        goto cleanup;
+      }
+
+    subject = g_object_new (ZEITGEIST_TYPE_SUBJECT, NULL);
+
+    g_variant_iter_loop (subject_data, "s", &str);
+    zeitgeist_subject_set_uri (subject, str[0] == '\0' ? NULL : str);
+
+    g_variant_iter_loop (subject_data, "s", &str);
+    zeitgeist_subject_set_interpretation (subject, str[0] == '\0' ? NULL : str);
+
+    g_variant_iter_loop (subject_data, "s", &str);
+    zeitgeist_subject_set_manifestation (subject, str[0] == '\0' ? NULL : str);
+
+    g_variant_iter_loop (subject_data, "s", &str);
+    zeitgeist_subject_set_origin (subject, str[0] == '\0' ? NULL : str);
+
+    g_variant_iter_loop (subject_data, "s", &str);
+    zeitgeist_subject_set_mimetype (subject, str[0] == '\0' ? NULL : str);
+
+    g_variant_iter_loop (subject_data, "s", &str);
+    zeitgeist_subject_set_text (subject, str[0] == '\0' ? NULL : str);
+
+    g_variant_iter_loop (subject_data, "s", &str);
+    zeitgeist_subject_set_storage (subject, str[0] == '\0' ? NULL : str);
+
+    zeitgeist_event_add_subject (result, subject);
+  }
+
+  /* Construct the event payload if necessary */
+  payload_size = g_variant_iter_n_children (payload_data);
+  if (payload_size > 0)
+  {
+    payload = g_byte_array_sized_new (payload_size);
+    while (g_variant_iter_next (payload_data, "y", &payload_byte))
+    {
+      // FIXME: Implicit guchar <-> uint8 conversion..?
+      g_byte_array_append (payload, &payload_byte, 1);
+    }
+    zeitgeist_event_set_payload (result, payload);
+  }
+
+  cleanup:
+    g_variant_iter_free (event_data);
+    g_variant_iter_free (subjects);
+    g_variant_iter_free (payload_data);
+    g_variant_unref (event);
+
+  return result;
+}
+
+/**
+ * zeitgeist_event_to_variant:
+ * @event: The #ZeitgeistEvent to serialize to a #GVariant
+ *
+ * Serialize a #ZeitgeistEvent to a #GVariant. The reverse operation
+ * is zeitgeist_event_new_from_variant().
+ *
+ * Returns: A floating reference to a #GVariant with signature
+ *          #ZEITGEIST_EVENT_VARIANT_SIGNATURE
+ */
+GVariant*
+zeitgeist_event_to_variant (ZeitgeistEvent *event)
+{
+  GVariant         *var;
+  GVariantBuilder   b;
+  gchar            *buf;
+  const gchar      *bif;
+  int               i, n_subjects;
+  ZeitgeistSubject *su;
+  GByteArray       *payload;
+
+  g_return_val_if_fail (ZEITGEIST_IS_EVENT (event), NULL);
+
+  g_variant_builder_init (&b, ZEITGEIST_EVENT_VARIANT_TYPE);
+
+  /* Build event data */
+  g_variant_builder_open (&b, G_VARIANT_TYPE ("as"));
+  if (zeitgeist_event_get_id (event) == 0)
+    buf = g_strdup ("");
+  else
+    buf = g_strdup_printf ("%"G_GUINT32_FORMAT, zeitgeist_event_get_id (event));
+  g_variant_builder_add (&b, "s", buf);
+  g_free (buf);
+  if (zeitgeist_event_get_timestamp (event) == 0)
+    buf = g_strdup ("");
+  else
+    buf = g_strdup_printf ("%"G_GINT64_FORMAT, zeitgeist_event_get_timestamp (event));
+  g_variant_builder_add (&b, "s", buf);
+  g_free (buf);
+  g_variant_builder_add (&b, "s", (bif = zeitgeist_event_get_interpretation(event), bif ? bif : ""));
+  g_variant_builder_add (&b, "s", (bif = zeitgeist_event_get_manifestation(event), bif ? bif : ""));
+  g_variant_builder_add (&b, "s", (bif = zeitgeist_event_get_actor (event), bif ? bif : ""));
+  g_variant_builder_close (&b);
+
+  /* Build subjects */
+  g_variant_builder_open (&b, G_VARIANT_TYPE ("aas"));
+  n_subjects = zeitgeist_event_num_subjects(event);
+  for (i = 0; i < n_subjects; i++)
+    {
+      su = zeitgeist_event_get_subject (event, i);
+      g_variant_builder_open (&b, G_VARIANT_TYPE ("as"));
+      g_variant_builder_add (&b, "s", (bif = zeitgeist_subject_get_uri(su), bif ? bif : ""));
+      g_variant_builder_add (&b, "s", (bif = zeitgeist_subject_get_interpretation(su), bif ? bif : ""));
+      g_variant_builder_add (&b, "s", (bif = zeitgeist_subject_get_manifestation(su), bif ? bif : ""));
+      g_variant_builder_add (&b, "s", (bif = zeitgeist_subject_get_origin(su), bif ? bif : ""));
+      g_variant_builder_add (&b, "s", (bif = zeitgeist_subject_get_mimetype(su), bif ? bif : ""));
+      g_variant_builder_add (&b, "s", (bif = zeitgeist_subject_get_text(su), bif ? bif : ""));
+      g_variant_builder_add (&b, "s", (bif = zeitgeist_subject_get_storage(su), bif ? bif : ""));
+      g_variant_builder_close (&b);
+    }
+  g_variant_builder_close (&b);
+
+  /* Build payload */
+  g_variant_builder_open (&b, G_VARIANT_TYPE ("ay"));
+  payload = zeitgeist_event_get_payload (event);
+  if (payload != NULL)
+    {
+      for (i = 0; i < payload->len; i++)
+        {
+          g_variant_builder_add (&b, "y", payload->data[i]);
+        }
+    }
+  g_variant_builder_close (&b);
+
+  return g_variant_builder_end (&b);
+}
+
+/**
+ * zeitgeist_events_to_variant:
+ * @events: A #GPtrArray of #ZeitgeistEvent<!-- -->s. If the events has
+ *          floating references they will be consumed. Furthermore the
+ *          reference on the #GPtrArray itself will also be stolen and its
+ *          @free_func set to %NULL
+ *
+ * Convert a set of #ZeitgeistEvent<-- -->s to a #GVariant with signature
+ * as an array of #ZEITGEIST_EVENT_VARIANT_SIGNATURE.
+ *
+ * Returns: A floating reference to a #GVariant as described above. Unless the
+ *          floating reference is consumed somewhere you must free it with
+ *          g_variant_unref().
+ */
+GVariant*
+zeitgeist_events_to_variant (GPtrArray *events)
+{
+  GVariantBuilder  b;
+  ZeitgeistEvent  *event;
+  GVariant        *vevent;
+  int              i;
+
+  g_return_val_if_fail (events != NULL, NULL);
+
+  g_variant_builder_init (&b,
+                          G_VARIANT_TYPE ("a"ZEITGEIST_EVENT_VARIANT_SIGNATURE));
+
+  for (i = 0; i < events->len; i++)
+    {
+      event = ZEITGEIST_EVENT (g_ptr_array_index (events, i));
+      g_object_ref_sink (event);
+      vevent = zeitgeist_event_to_variant (event);
+      g_variant_builder_add_value (&b, vevent);
+      g_object_unref (event);
+    }
+
+  /* We need to unset the free func because we can't risk double unreffing
+   * the events held in it, quite ugly but there is no other way */
+  g_ptr_array_set_free_func (events, NULL);
+  g_ptr_array_unref (events);
+
+  return g_variant_builder_end (&b);
+}
+
+/**
+ * zeitgeist_events_from_variant:
+ * @events: A #GVariant  with signature as an array of
+ *          #ZEITGEIST_EVENT_VARIANT_SIGNATURE. If @event is floating this
+ *          method will consume the floating reference.
+ *
+ * Returns: A reference to a #GPtrArray of #ZeitgeistEvent<!-- -->s.
+ *          All the events will be floating references, and the
+ *          #GPtrArray<!-- -->'s @free_func will be set to g_object_unref().
+ */
+GPtrArray*
+zeitgeist_events_from_variant (GVariant *events)
+{
+  GPtrArray      *result;
+  GVariant       *vevent;
+  ZeitgeistEvent *event;
+  int             i, n_events;
+
+  g_return_val_if_fail (events != NULL, NULL);
+
+  g_variant_ref_sink (events);
+
+  n_events = g_variant_n_children (events);
+  result = g_ptr_array_sized_new (n_events);
+  g_ptr_array_set_free_func (result, (GDestroyNotify) g_object_unref);
+
+  for (i = 0; i < n_events; i++)
+    {
+      vevent = g_variant_get_child_value (events, i);
+      event = zeitgeist_event_new_from_variant (vevent); // vevent freed
+      g_ptr_array_add (result, event);
+    }
+
+  g_variant_unref (events);
+
+  return result;
+}
+
+/**
+ * zeitgeist_events_from_valist:
+ * @events: a #va_list of #ZeitgeistEvent<!-- -->s.
+ *
+ * Utility function to convert a #va_list of #ZeitgeistEvent<!-- --> into a
+ * #GPtrArray containing these events.
+ *
+ * Returns: A #GPtrArray of #ZeitgeistEvent<!-- -->s. The events are not further
+ *          reffed and will still be floating if they where so when you called
+ *          this method. The pointer array will have its @free_func set to
+ *          g_object_unref()
+ */
+GPtrArray*
+zeitgeist_events_from_valist (va_list events)
+{
+  ZeitgeistEvent *event = NULL;
+  GPtrArray *result;
+
+  result = g_ptr_array_new_with_free_func ((GDestroyNotify) g_object_unref);
+
+  event = va_arg (events, ZeitgeistEvent*);
+  while (event != NULL)
+    {
+      g_return_val_if_fail (ZEITGEIST_IS_EVENT (event), NULL);
+      g_ptr_array_add (result, event);
+      event = va_arg (events, ZeitgeistEvent*);
+    }
+
+  return result;
+}
