@@ -3,15 +3,16 @@
  *
  * This library is free software; you can redistribute it and/or modify
  * it under the terms of the GNU Lesser General Public License
- * version 3.0 as published by the Free Software Foundation.
+ * as published by the Free Software Foundation; either
+ * version 2.1 of the License, or (at your option) any later version.
  *
  * This library is distributed in the hope that it will be useful,
  * but WITHOUT ANY WARRANTY; without even the implied warranty of
  * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
- * GNU Lesser General Public License version 3.0 for more details.
+ * GNU Lesser General Public License for more details.
  *
  * You should have received a copy of the GNU Lesser General Public
- * License along with this library. If not, see
+ * License along with this library; if not, see
  * <http://www.gnu.org/licenses/>.
  *
  * Authored by: Michal Hruby <michal.mhr@gmail.com>
@@ -111,7 +112,7 @@ zeitgeist_symbol_get_parents (const gchar *symbol)
   for (iter = info->parents; iter; iter = iter->next)
   {
     result = g_list_prepend (result,
-                             g_quark_to_string (GPOINTER_TO_INT (iter->data)));
+                             (gchar*) g_quark_to_string (GPOINTER_TO_INT (iter->data)));
   }
 
   return g_list_reverse (result);
@@ -143,7 +144,7 @@ zeitgeist_symbol_get_children (const gchar *symbol)
   for (iter = info->children; iter; iter = iter->next)
   {
     result = g_list_prepend (result,
-                             g_quark_to_string (GPOINTER_TO_INT (iter->data)));
+                             (gchar*) g_quark_to_string (GPOINTER_TO_INT (iter->data)));
   }
 
   return g_list_reverse (result);
@@ -176,7 +177,7 @@ zeitgeist_symbol_get_all_children (const gchar *symbol)
   for (iter = info->all_children; iter; iter = iter->next)
   {
     result = g_list_prepend (result,
-                             g_quark_to_string (GPOINTER_TO_INT (iter->data)));
+                             (gchar*) g_quark_to_string (GPOINTER_TO_INT (iter->data)));
   }
 
   return g_list_reverse (result);
